import { BrowserWindow, } from 'electron';
import { wm_create_conf, wm_move_conf, wm_resize_conf } from './../types/wm';
import { toP, screenSize } from './utils';

export type WindowData = {
  win: BrowserWindow;
  conf: wm_create_conf;
  wid: string;
  follows?: string;
  followStuff?: any;
}

let wdata: { [key: string]: WindowData } = {};

const wm = {
  create: function (c: wm_create_conf) {
    let wid = Array.from({ length: 32 }, () => '0123456789abcdef'[Math.floor(Math.random() * 16)]).join('');
    while (wdata[wid]) {
      console.warn('[wm] create: got a wid duplicate');
      wid = Array.from({ length: 32 }, () => '0123456789abcdef'[Math.floor(Math.random() * 16)]).join('');
    }
    console.log('[wm] create', wid);
    const { width: sw, height: sh } = screenSize();
    const win = new BrowserWindow({
      ...c.customProps,
      width: Math.round(c.whp ? toP(sw, c.w) : c.w),
      height: Math.round(c.whp ? toP(sh, c.h) : c.h),
      x: c.x ? Math.round(c.xyp ? toP(sw, c.x) : c.x) : undefined,
      y: c.y ? Math.round(c.xyp ? toP(sh, c.y) : c.y) : undefined,
      frame: !c.noBorder,
      transparent: c.noBackground,
      webPreferences: {
        nodeIntegration: true,
        contextIsolation: false,
      }
    });

    const wobj = { win: win, conf: c, wid: wid } as WindowData;
    wdata[wid] = wobj;
    win.setMenuBarVisibility(false); // tell me WHY should i NOT make this the default
    win.on('close', () => { delete wdata[wid]; this.follow.stop(wid) });
    if (c.onCreate) c.onCreate(wobj);

    if (typeof c.onMinimize === 'function') win.on('minimize', () => c.onMinimize!(wobj));
    if (typeof c.onMaximize === 'function') win.on('maximize', () => c.onMaximize!(wobj));
    if (typeof c.onRestore === 'function') win.on('restore', () => c.onRestore!(wobj));
    if (typeof c.onFocus === 'function') win.on('focus', () => c.onFocus!(wobj));
    if (typeof c.onShow === 'function') win.on('show', () => c.onShow!(wobj));
    if (typeof c.onUnfocus === 'function') win.on('blur', () => c.onUnfocus!(wobj));
    if (typeof c.onClose === 'function') win.on('close', () => c.onClose!(wobj));
    if (typeof c.onMove === 'function') win.on('move', () => c.onMove!(wobj));
    if (typeof c.onResize === 'function') win.on('resize', () => c.onResize!(wobj));

    return wid // just so you can reference it later on
  },
  destroy: async function (id: string) {
    if (!wdata[id]) return
    console.log('[wm] destroy', id);
    const win = wdata[id];
    if (typeof win.conf.onDestroy === 'function') await win.conf.onDestroy!(win);
    win.win.destroy();
    delete wdata[id];
  },
  move: function (c: wm_move_conf) {
    if (!wdata[c.id]) return
    console.log('[wm] move', c.id);
    const win = wdata[c.id];
    const { x: cwx, y: cwy } = win.win.getBounds();
    const { width: sw, height: sh } = screenSize();
    let wb = win.win.getBounds();
  
    const sX = wb.x;
    const sY = wb.y;
    c.x = c.x === undefined ? sX : Math.round(c.x);
    c.y = c.y === undefined ? sY : Math.round(c.y);
  
    const duration = c.duration || 500;
    const tick = c.tick || 1000 / 30;
    const totalSteps = Math.floor(duration / tick);

    //                       is relative?              add current window's x/y
    //                       |   relative              |           |    absolute
    //                       |   | is %?               |           |    | is %?
    //                       |   | |     %             |     px    |    | |     %              px
    let targetX = Math.round(c.r ? c.p ? toP(sw, c.x) + cwx : c.x + cwx : c.p ? toP(sw, c.x) : c.x);
    let targetY = Math.round(c.r ? c.p ? toP(sh, c.y) + cwy : c.y + cwy : c.p ? toP(sh, c.y) : c.y);
  
    if (c.fromCenter) {
      wb = win.win.getBounds();
      targetX = targetX - wb.width * 0.5;
      targetY = targetY - wb.height * 0.5;
    }
  
    const animate = () => {
      let currentStep = 0;
      const startX = sX;
      const startY = sY;
  
      const step = () => {
        currentStep++;
        const t = currentStep / totalSteps;
        const eT = c.ease ? c.ease(t) : t;
  
        const newX = startX + (targetX - startX) * eT;
        const newY = startY + (targetY - startY) * eT;
  
        win.win.setPosition(Math.round(newX), Math.round(newY));
  
        if (currentStep < totalSteps) {
          setTimeout(step, tick);
        } else {
          win.win.setPosition(Math.round(targetX), Math.round(targetY));
        }
      };
  
      step();
    };
  
    if (c.smooth) {
      animate();
    } else {
      win.win.setPosition(Math.round(targetX), Math.round(targetY));
    }
  },
  resize: function (c: wm_resize_conf) {
    if (!wdata[c.id]) return
    console.log('[wm] resize', c.id);
    const win = wdata[c.id];
    if (!win || !win.win) {
      console.error('Window data not found for id:', c.id);
      return;
    }
    const wb = win.win.getBounds();
    const sW = wb.width, sH = wb.height;
    c.w = c.w === undefined ? sW : Math.round(c.w);
    c.h = c.h === undefined ? sH : Math.round(c.h);
    const duration = c.duration || 500;
    const tick = c.tick || 1000 / 30;
    const totalSteps = Math.floor(duration / tick);
    const { width: sw, height: sh } = screenSize();
    const targetW = Math.round(toP(c.p ? sw : 1, c.w));
    const targetH = Math.round(toP(c.p ? sh : 1, c.h));
  
    const centerX = Math.round(wb.x + wb.width / 2);
    const centerY = Math.round(wb.y + wb.height / 2);
  
    const getAnchoredPosition = (curWidth: number, curHeight: number) => {
      let newX = wb.x;
      let newY = wb.y;
      switch (c.anchor) {
        case 'top':
          newY = wb.y;
          break;
        case 'bottom':
          newY = wb.y + wb.height - curHeight;
          break;
        case 'left':
          newX = wb.x;
          break;
        case 'right':
          newX = wb.x + wb.width - curWidth;
          break;
        default:
          newX = centerX - curWidth / 2;
          newY = centerY - curHeight / 2;
      }
      return { newX, newY };
    };
  
    const animate = () => {
      let currentStep = 0;
      const startW = wb.width;
      const startH = wb.height;
      const step = () => {
        currentStep++;
        const t = Math.min(currentStep / totalSteps, 1);
        const eT = c.ease ? c.ease(t) : t;
        const newW = startW + (targetW - startW) * eT;
        const newH = startH + (targetH - startH) * eT;
        const pos = getAnchoredPosition(Math.round(newW), Math.round(newH));
        win.win.setBounds({
          x: pos.newX,
          y: pos.newY,
          width: Math.round(newW),
          height: Math.round(newH)
        });
        if (currentStep < totalSteps) {
          setTimeout(step, tick);
        } else {
          const finalPos = getAnchoredPosition(Math.round(targetW), Math.round(targetH));
          win.win.setBounds({
            x: finalPos.newX,
            y: finalPos.newY,
            width: Math.round(targetW),
            height: Math.round(targetH)
          });
        }
      };
      step();
    };
    if (c.smooth) {
      animate();
    } else {
      const finalPos = getAnchoredPosition(Math.round(targetW), Math.round(targetH));
      win.win.setBounds({
        x: finalPos.newX,
        y: finalPos.newY,
        width: Math.round(targetW),
        height: Math.round(targetH)
      });
    }
  },  
  eval: async function (id: string, code: string) {
    if (!wdata[id]) return
    console.log('[wm] eval', id);
    try {
      return await wdata[id].win.webContents.executeJavaScript(code);
    } catch (error) {
      console.error('[wm] eval e:', error);
    }
  },
  follow: {
    INTERNAL_follow: function (id: string) {
      if (!wdata[id]) return
      const leadWin = wdata[wdata[id].follows!].win;
      const followWin = wdata[id].win;
      const leadBounds = leadWin.getBounds();

      const newX = leadBounds.x + wdata[id].followStuff.offsetX;
      const newY = leadBounds.y + wdata[id].followStuff.offsetY;

      followWin.setPosition(newX, newY);
      followWin.show();
    },
    INTERNAL_followfocus: function (id: string) {
      if (wdata[id]) wdata[id].win.show();
    },
    start: function (id: string, follows: string) {
      if (!wdata[id] || !wdata[follows]) return
      console.log('[wm]', id, 'follows', follows);
      wdata[id].follows = follows;

      const fb = wdata[follows].win.getBounds();
      const lb = wdata[id].win.getBounds();
      wdata[id].followStuff = {
        offsetX: lb.x - fb.x,
        offsetY: lb.y - fb.y
      };

      wdata[follows].win.addListener('move', this.INTERNAL_follow.bind(this, id));
      wdata[follows].win.addListener('focus', this.INTERNAL_followfocus.bind(this, id));
      wdata[follows].win.addListener('show', this.INTERNAL_followfocus.bind(this, id));
    },    
    stop: function (id: string) {
      if (!wdata[id]) return
      wdata[wdata[id].follows!].win.removeListener('move', this.INTERNAL_follow.bind(this, id));
      wdata[wdata[id].follows!].win.removeListener('focus', this.INTERNAL_followfocus.bind(this, id));
      wdata[id].followStuff = {};
    }
  }
};

export default wm;
export { wdata, wm };